/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*global document, window, WebHelperClient */

/**
 * IME Keyboard application.
 * This application captures keyboard click events, parse them if necessary
 * and sends to system service middleware (WebHelperClient).
 *
 * @module app
 * @requires {@link web-helper-client}
 * @namespace app
 */

(function app() {

    'use strict';

    /**
     * Defines HTML id attributes for different keyboard layouts.
     *
     * @memberof app
     * @private
     * @const {object}
     */
    var LAYOUTS = {
            PHONE_NUMBER: 'phonenumber',
            NUMBERS: 'number',
            NUMBERS_SYMBOLS: 'numsys1',
            ENGLISH_LOWERCASE: 'english-lowercase',
            ENGLISH_UPPERCASE: 'english-uppercase'
        },

        /**
         * Current layout additional HTML class name.
         *
         * @memberof app
         * @private
         * @const {string}
         */
        CURRENT_LAYOUT_CLASSNAME = 'visible',

        /**
         * Keyboard height in pixels.
         *
         * @memberof app
         * @private
         * @const {number}
         */
        KEYBOARD_HEIGHT = 200,

        /**
         * Accessing next key value timeout in milliseconds.
         *
         * @memberof app
         * @private
         * @const {number}
         */
        KEY_LASTING_TIME = 500,

        /**
         * Stores keyboard layouts containers HTML references.
         *
         * @memberof app
         * @private
         * @type {HTMLElement[]}
         */
        layoutContainers = [],

        /**
         * Stores references to HTML elements switching keyboard layouts.
         *
         * @memberof app
         * @private
         * @type {HTMLElement[]}
         */
        layoutSwitches = [],

        /**
         * Stores last clicked key information.
         *
         * @memberof app
         * @private
         * @type {object}
         * @property {HTMLElement} htmlElement Pressed button HTML element.
         * @property {number} age Time since click in milliseconds.
         * @property {number} count Element press count.
         */
        lastPressedKey = {
            htmlElement: null,
            timestamp: 0,
            count: 0
        },

        /**
         * System service middleware configuration object.
         *
         * @memberof app
         * @private
         * @type {object}
         */
        webHelperClientHandler = {};

    /**
     * Gets elements with querySelectorAll and returns them in array.
     *
     * @memberof app
     * @private
     * @param {string} selector DOM Selector.
     * @param {HTMLElement} [element=document] Search scope.
     * @returns {HTMLElement[]}
     */
    function getElementsInArray(selector, element) {
        return Array.prototype.slice.apply(
            (element || document).querySelectorAll(selector)
        );
    }

    /**
     * Changes keyboard layout to the one provided with clicked element
     * data-layout attribute.
     *
     * @memberof app
     * @private
     * @param {MouseEvent} event
     * @param {string} layout Layout indicator.
     */
    function onInputModeChanged(event, layout) {
        var to = typeof layout === 'string' ?
                layout : event.target.getAttribute('data-layout');

        if (event) {
            event.preventDefault();
        }

        document.querySelector('.' + CURRENT_LAYOUT_CLASSNAME)
            .classList.remove(CURRENT_LAYOUT_CLASSNAME);
        document.getElementById(to).classList.add(CURRENT_LAYOUT_CLASSNAME);
    }

    /**
     * Sends provided charcode to system service middleware.
     *
     * @memberof app
     * @private
     * @param {number} key
     */
    function keyCode(key) {
        WebHelperClient.sendKeyEvent(key);
    }

    /**
     * Sends key event for each character from specified string.
     *
     * @memberof app
     * @private
     * @param {string} string
     */
    function sendKeyEventsFromString(string) {
        var i = 0,
            length = string.length;

        for (i = 0; i < length; i += 1) {
            WebHelperClient.sendKeyEvent(string.charCodeAt(i));
        }
    }

    /**
     * Handles key press.
     *
     * @memberof app
     * @private
     * @param {MouseEvent} event
     */
    function onKeyPressed(event) {
        var keyCharacters = event.target.innerText,
            charsCount = keyCharacters.length;

        event.preventDefault();

        if (keyCharacters.length > 1 && lastPressedKey.htmlElement &&
                lastPressedKey.htmlElement.innerText === keyCharacters &&
                Date.now() - lastPressedKey.timestamp < KEY_LASTING_TIME) {

            keyCode(WebHelperClient.Keycode.BACKSPACE);
            lastPressedKey.count = lastPressedKey.count % charsCount + 1;
        } else {
            lastPressedKey.count = 1;
        }

        lastPressedKey.htmlElement = event.target;
        lastPressedKey.timestamp = Date.now();

        sendKeyEventsFromString(keyCharacters[lastPressedKey.count - 1]);
    }

    /**
     * Handles function key press.
     * Calls keyCode function with charcode from provided element attribute.
     *
     * @memberof app
     * @private
     * @param {MouseEvent} event
     */
    function onFunctionKey(event) {
        event.preventDefault();

        keyCode(WebHelperClient.Keycode[
            event.target.getAttribute('data-fn')
        ]);
    }

    /**
     * Handles keys providing characters strings.
     *
     * @memberof app
     * @private
     * @param {MouseEvent} event
     */
    function onStringKey(event) {
        event.preventDefault();

        sendKeyEventsFromString(event.target.getAttribute('data-str'));
    }

    /**
     * Keyboard system service middleware configuration.
     *
     * @memberof app
     * @private
     * @type {object}
     */
    webHelperClientHandler = {

        /**
         * Handles initialization signal,
         * used when the Web IME framework initialization is finished.
         *
         */
        onInit: function init() {
            WebHelperClient.log('ON INIT');
        },

        /**
         * Handles virtual keyboard layout signal set,
         * used to set the virtual keyboard layout.
         *
         * @param {string} layout
         */
        onSetLayout: function onSetLayout(layout) {
            var nextMode = null;

            if (layout === LAYOUTS.PHONE_NUMBER) {
                nextMode = LAYOUTS.NUMBERS_SYMBOLS;
            } else if (layout === LAYOUTS.NUMBERS) {
                nextMode = LAYOUTS.NUMBERS_SYMBOLS;
            } else {
                nextMode = LAYOUTS.ENGLISH_LOWERCASE;
            }
            onInputModeChanged(null, nextMode);
        },

        /**
         * Handles keyboard show signal.
         * Sets keyboard dimensions.
         */
        onShow: function show() {
            var screenWidth = window.screen.width,
                screenHeight = window.screen.height;

            WebHelperClient.setKeyboardSizes(
                screenWidth,
                KEYBOARD_HEIGHT,
                screenHeight,
                KEYBOARD_HEIGHT
            );
        }
    };

    /**
     * Binds events to keyboard keys.
     *
     * @memberof app
     * @private
     */
    function bindEvents() {
        layoutSwitches.forEach(function bindLayoutSwitches(elem) {
            elem.addEventListener('click', onInputModeChanged);
        });

        getElementsInArray('[data-fn]')
            .forEach(function bindFunctionKeys(elem) {
                elem.addEventListener('click', onFunctionKey);
            });

        getElementsInArray(
            '.key:not([data-fn]):not([data-layout]):not([data-str])'
        ).forEach(function bindCharKeys(elem) {
            elem.addEventListener('click', onKeyPressed);
        });

        getElementsInArray('[data-str]').forEach(function bindStringKeys(elem) {
            elem.addEventListener('click', onStringKey);
        });
    }

    /**
     * Initializes variables and calls startup function.
     *
     * @memberof app
     * @private
     */
    function init() {
        layoutContainers = getElementsInArray('.layout');
        layoutSwitches = getElementsInArray('[data-layout]');
        bindEvents();

        // show first layout

        layoutContainers[0].classList.add(CURRENT_LAYOUT_CLASSNAME);
    }

    // initialize system service middleware

    WebHelperClient.initialize(webHelperClientHandler);

    init();

}());
